/**
 * @file block.cs 
 * Classes to represent and evaluate C# statement blocks and statements.
 *
 * (C) 2004 Martin Pitt <martin@piware.de>
 *
 * This software is distributed under the terms and conditions of the GNU
 * General Public License (see file COPYING)
 */

using System;
using System.IO;
using System.Collections;

namespace CCM {

/// Common interface of all statements.
public abstract class Statement {
    /// Execute statement. 
    abstract public void execute();
}

/// Statement block (list of Statements).
public class Block : Statement {

    /// Create empty block as a subblock of parent (null for toplevel blocks).
    public Block( Block parent ) {
	stm = new ArrayList( 5 );
	vars = new Hashtable( 5 );
	this.parent = parent;
    }

    /// Return the block's parent (null for toplevel blocks).
    public Block getParent() { return parent; }

    /// Return the block that is currently executed.
    public static Block getCurrentBlock() {
	return currentBlock;
    }

    // statement handling

    /// Add the given statement to the end of the block.
    public void addStm( Statement s ) {
	if( s != null )
	    stm.Add( s );
    }

    /// Add given list of Statements to the end of the block.
    public void addStm( IList sl ) {
	if( sl != null )
	    foreach( Statement s in sl )
		addStm( s );
    }

    public override void execute() {
	Block oldBlock = currentBlock;
	currentBlock = this;

	foreach( Statement s in stm ) {
	    if( Block.breakFlag || Block.continueFlag || Method.returning() )
		break;
	    s.execute();
	}

	currentBlock = oldBlock;
    }

    // variable handling

    /// Add given local variable to the block.
    public void addVar( Variable v ) {
	vars.Add( v.name, v );
    }

    /**
     * Return the local variable with given name. Also looks for variables in
     * parent blocks.
     */
    public Variable getVar( string name ) {
	if( vars.ContainsKey( name ) )
	    return (Variable) vars[name];
	if(  parent != null )
	    return parent.getVar( name );
	else
	    return null;
    }

#if DEBUG
    // debugging

    /// Write the value of all variables to the standard output (for debugging
    /// purposes).
    public void dumpVars() {
	foreach( DictionaryEntry i in vars )
	    Console.WriteLine( i.Value );
    }
#endif

    /// Local variables (string -> Variable).
    Hashtable vars; 

    /// Contained Statements.
    IList stm; 

    /// Parent block (null for toplevel blocks).
    Block parent;

    /// Global pointer to get the current variable scope; handled by execute().
    static Block currentBlock = null;

    /// If true, the current loop must be break'ed.
    public static bool breakFlag = false;
    
    /// If true, the current loop must be continue'd.
    public static bool continueFlag = false;
}

/// Statement that just evaluates an expression.
public class ExprStm : Statement {
    /// Create expression statement.
    public ExprStm( Expr e ) {
	expr = e;
    }

    public override void execute() {
	expr.eval();
    }

    Expr expr;
}

/// If-Then-Else statement.
public class IfStm: Statement {
    /// Create "if" statement with given condition, then- and else statement.
    public IfStm( BoolExpr cond, Statement then, Statement els ) {
	this.cond = cond;
	this.then = then;
	this.els = els;
    }

    public override void execute() {
	if( (bool) cond.eval() ) {
	    if( then != null )
		then.execute();
	} else {
	    if( els != null )
		els.execute();
	}
    }

    BoolExpr cond;
    Statement then, els;
}

/// While loop statement.
public class WhileStm: Statement {
    /// Create "while" statement with given condition and body.
    public WhileStm( BoolExpr cond, Statement body ) {
	this.cond = cond;
	this.body = body;
    }

    public override void execute() {
	if( body == null )
	    while( (bool) cond.eval() ) {}
	else {
	    Block.breakFlag = Block.continueFlag = false;
	    while( !Method.returning() && (bool) cond.eval() ) {
		body.execute(); 
		if( Block.breakFlag ) {
		    Block.breakFlag = false;
		    break;
		}
		if( Block.continueFlag )
		    Block.continueFlag = false;
	    }
	}
    }

    BoolExpr cond;
    Statement body;
}

/// For loop statement without initializer (this is already handled by the parser).
public class ForStm: Statement {
    /// Create "for" statement with given condition, iterator Statements and body.
    public ForStm( BoolExpr cond, IList iters, Statement body ) {
	this.cond = cond;
	this.iters = iters;
	this.body = body;
    }

    public override void execute() {
	if( body == null ) {
	    while( (bool) cond.eval() )
		foreach( Statement i in iters )
		    i.execute();
	} else {
	    Block.breakFlag = Block.continueFlag = false;
	    while(!Method.returning() && (bool) cond.eval() ) {
		body.execute(); 
		if( Block.breakFlag ) {
		    Block.breakFlag = false;
		    break;
		}
		if( Block.continueFlag )
		    Block.continueFlag = false;
		foreach( Statement i in iters )
		    i.execute();
	    }
	}
    }

    BoolExpr cond;
    IList iters; // of Statement
    Statement body;
}

/// Do loop statement.
public class DoStm: Statement {
    /// Create "do" statement with given condition and body.
    public DoStm( BoolExpr cond, Statement body ) {
	this.cond = cond;
	this.body = body;
    }

    public override void execute() {
	if( body == null )
	    while( (bool) cond.eval() ) {}
	else {
	    Block.breakFlag = Block.continueFlag = false;
	    do { 
		body.execute(); 
		if( Block.breakFlag ) {
                    Block.breakFlag = false;
                    break;
                }
                if( Block.continueFlag )
                    Block.continueFlag = false;

	    } while( !Method.returning() && (bool) cond.eval() );
	}
    }

    BoolExpr cond;
    Statement body;
}

/// Break statement.
public class BreakStm: Statement {
    /// Create break statement.
    public BreakStm() {}
    
    public override void execute() {
	Block.breakFlag = true;
    }
}

/// Continue statement.
public class ContinueStm: Statement {
    /// Create continue statement.
    public ContinueStm() {}
    
    public override void execute() {
	Block.continueFlag = true;
    }
}

/// Return statement.
public class ReturnStm: Statement {
    /** 
     * Create return statement.
     * @param val Return value; use null for void methods
     */
    public ReturnStm( Expr val ) { this.val = val; }
    
    public override void execute() {
	Method.setReturn( (val == null) ? null : val.eval() );
    }

    Expr val;
}

/// Throw statement.
public class ThrowStm : Statement {
    /// Create throw statement with given expression as exception.
    public ThrowStm( Expr e ) {
	exception = e;
    }

    public override void execute() {
	Console.WriteLine( "Component throws exception " + exception +
		"\nTerminating cs2ccmb." );
	throw new Exception( "Component throws exception" );
    }

    Expr exception;
}

/// Statement that writes an expression to the standard output.
public class WriteStm : Statement {
    /**
     * Create write statement.
     * @param e The expression whose value is to be printed
     * @param eol When true, a newline is appended to the output
     */
    public WriteStm( Expr e, bool eol ) {
	expr = e;
	this.eol = eol;
    }

    public override void execute() {
	if( eol )
	    Console.WriteLine( expr.eval().ToString() );
	else
	    Console.Write( expr.eval().ToString() );
    }

    Expr expr;
    bool eol;
}

} // closing namespace
