/**
 * @file class.cs 
 * Classes to represent classes and methods.
 *
 * (C) 2004 Martin Pitt <martin@piware.de>
 *
 * This software is distributed under the terms and conditions of the GNU
 * General Public License (see file COPYING)
 */

using System;
using System.IO;
using System.Collections;
using System.Text.RegularExpressions;

namespace CCM {

/**
 * Represents a method. This class implements IEnumerable to iterate over all
 * possible call arguments.
 */
public class Method : IEnumerable {
    /**
     * Create new method.
     * @param name Method name
     * @param retType Return type (null for void)
     * @param pars Method parameters (list of VariableDecl)
     * @param attrs List of Attributes
     * @param _static true if this is a static method
     */
    public Method( string name, Type retType, IList pars, List attrs, bool _static) {
	this.name = name;
	this.pars = pars;
	this.retType = retType;
	this.body = null;
	this.attrs = attrs;
	this._static = _static;
    }

    /**
     * Set the body of the class.  This can not be done in the constructor,
     * because the parameters are already needed when parsing the body block.
     */
    public void setBody( Block body ) { this.body = body; }

    /// Return the method's name.
    public string getName() { return name; }

    /// Return the number of parameters.
    public int numParams() { 
	return pars.Count;
    }

    /// Return whether this is a static method.
    public bool isStatic() { return _static; }

    /// Get the declaration of the parameter with given name.
    public VariableDecl getParam( string name ) {
	foreach( VariableDecl v in pars )
	    if( v.name == name )
		return v;
	return null;
    }

    /// Get the return type (null means void).
    public Type returnType() {
	return retType;
    }

    /// Returns the first attribute that has the given name.
    public Attribute getAttribute( string name ) {
	if( attrs != null )
	    foreach( Attribute a in attrs )
		if( a.name == name )
		    return a;
	return null;
    }

    /// Return whether this method is callable with arguments args (which must
    /// be either proper values or Expr derivatives).
    public bool compatibleArgs( IList args ) {
    //TODO: respect arrays
	Type t;
	Type objectType = (new Object() ).GetType();

	if( args.Count != pars.Count )
	    return false;
	for( int i = 0; i < args.Count; ++i ) {
	    Type paramType = ( (VariableDecl) pars[i] ).type;
	    if( paramType == objectType )
		continue;
	    if( args[i] is Expr ) 
		t = ( (Expr) args[i] ).type();
	    else
		t = args[i].GetType();
	    if( t != paramType )
		return false;
	}
	return true;
    }

    /// Returns wheter this method is ambiguous to the given one (same name and
    /// parameters).
    public bool isAmbiguous( Method m ) {
	if( m.name != name || pars.Count != m.pars.Count )
	    return false;
	for( int i = 0; i < pars.Count; ++i )
	    if( ( (VariableDecl) pars[i] ).type != ( (VariableDecl) m.pars[i] ).type )
		return false;
	return true;
    }

    /// Execute the method with given parameters (which must be proper values). 
    public Object execute( IList paramValues ) {
	if( body == null )
	    throw new Exception( "Method.execute(): no body set" );
	// since this is already checked at the parsing stage, this could be
	// dropped for efficiency
	if( !compatibleArgs( paramValues ) )
	    throw new Exception( "Method.execute(): called " + name + "(" +
		    pars + ") with incompatible arguments (" + paramValues +")" );

	calledMethods.Push( new MethodCall( this, paramValues ) );

	returnFlag = false;
	returnVal = null;

#if DEBUG
	Console.WriteLine( "--- begin method " + name + ", var dump: -----" );
	body.dumpVars();
#endif

	body.execute();

#if DEBUG
	Console.WriteLine( "--- end method " + name + ", var dump: -----" );
	body.dumpVars();

	if( returnVal != null )
	    Console.WriteLine( "Returns: " + returnVal );
#endif

	calledMethods.Pop();

	if( retType != null && returnFlag == false )
	    throw new Exception( "Method.execute(): no value returned in non-void method" );

	Object result = returnVal;
	returnFlag = false;
	returnVal = null;

	return result;
    }

    public override string ToString() {
	string result;
	if( attrs != null && attrs.Count > 0 )
	    result = "[" + attrs + "] ";
	else
	    result = "";
	if( _static )
	    result += "static ";
	result += ( retType==null ? "void" : retType.ToString() ) + " " 
	    + name + "(" + pars + ")";
	return result;
    }

    string name;
    Type retType;
    IList pars; // of VariableDecl
    Block body;
    List attrs; // of Attribute
    bool _static;

    // return statement handling
    
    static bool returnFlag = false;
    static Object returnVal = null;

    /// Returns true if a return statement was executed (thus the statements
    /// following it must not be executed).
    public static bool returning() { return returnFlag; }

    /// Set the returning flag and return the given value.
    public static void setReturn( Object val ) {
	returnFlag = true;
	returnVal = val;
    }

    /// Get an iterator over the method parameters.
    public IEnumerator GetEnumerator() {
	return new ListIterator( pars );
    }

    /// Get an iterator over the return values.
    public IEnumerator getReturnEnumerator() {
	if( retType == false.GetType() )
	    return new BoolIterator();

	if( retType == 0.GetType() )
	    return new IntIterator( getAttribute( "Range" ) );
	if( retType == (0.0d).GetType() )
	    return new DoubleIterator( getAttribute( "Range" ) );

	throw new Exception( "Method.getReturnEnumerator(): cannot iterate over " + retType );
    }

    /**
     * Stack of currently called methods; it is here in class Method to
     * keep it private (so only Method.execute() can modify it).
     */
    struct MethodCall {
	/**
	 * Creates a method stack entry. 
	 * @param m Called method
	 * @param args List of actual parameter values; they are converted to a proper Variable
	 *        object to be compatible to Block and Class variables.
	 */
	public MethodCall( Method m, IList args ) { 
	    method = m;
	    vars = new ArrayList();
	    for( int i = 0; i < args.Count; ++i ) {
		VariableDecl p = (VariableDecl) m.pars[i];
		vars.Add( new Variable( p.name, p.type, p.stubClass, p.array, null, false, args[i] ) );
	    }
	}

	public Method method;
	public IList vars; // of Variable
    }

    /// Stack of currently called methods (of MethodCall).
    static Stack calledMethods = new Stack();

    /// Get method parameter value by name (for resolving in expressions).
    public static Variable getCallParam( string name ) {
	MethodCall call = (MethodCall) calledMethods.Peek();
	foreach( Variable v in call.vars )
	    if( v.name == name )
		return v;
	return null;
    }
}

/// Represents a class.
public class Class {
    /**
     * Create a new class.
     * @param name Class name.
     * @param attrs List of Attributes of this class
     * @param parent Name of the class this class is derived from.
     */
    public Class( string name, List attrs, string parent ) {
	this.name = name;
	methods = new ArrayList();
	ctors = new ArrayList();
	fields = new Hashtable();
	this.attrs = attrs;
	this.parent = parent;
	initStatements = null;
	isStub = false;

	if( parent == "Comp" )
	    addCCMCompMembers();
	else if( parent == "Stub" ) {
	    addCCMStubMembers();
	    isStub = true;
	} 
    }

    /// Instantiate an object of this class by sharing methods, ctors and
    /// attributes and deep-copying the fields.
    public Class( Class c ) {
	name = c.name;
	methods = c.methods;
	ctors= c.ctors;
	attrs = c.attrs;
	parent = c.parent;
	isStub = c.isStub;

	fields = new Hashtable();
	foreach( DictionaryEntry i in c.fields ) {
	    Variable f = (Variable) i.Value;
	    fields.Add( f.name, new Variable( f ) );
	}
    }

    /// Return the class name.
    public string getName() { return name; }

    /// Return the class name this class is derived from.
    public string parentClass() { return parent; }

    //
    // member handling
    //

    /// Add given method as a constructor.
    public void addConstructor( Method m ) {
	ctors.Add( m );
    }

    /**
     * Add given method. 
     * @return true on success or false if a method with the same name and
     *         parameters already exists.
     */
    public bool addMethod( Method method ) {
	foreach( Method m in methods )
	    if( method.isAmbiguous( m ) )
		return false;
	methods.Add( method );
	return true;
    }

    /**
     * Delete the given method from this class. This is necessary for parsing
     * when the parser discovers that a declared method has no body.
     */
    public void delMethod( Method m ) {
	methods.Remove( m );
    }

    /// Returns the method that matches given name and parameters (proper
    /// values, or Expr derivatives).
    public Method getMethod( string name, IList args ) {
	foreach( Method m in methods )
	    if( name == m.getName() && m.compatibleArgs( args ) )
		return m;
	return null;
    }

    /// Add given variable as class field.
    public void addField( Variable f ) {
	fields[f.name] = f;
    }

    /// Return field with given name (or null if such a field does not exist).
    public Variable getField( string name ) {
	return (Variable) fields[name];
    }

    /**
     * Add statement that must be executed in every constructor. Used to
     * initialize fields.
     */
    public void addInitStatement( Statement s ) {
	if( initStatements == null )
	    initStatements = new ArrayList();
	initStatements.Add( s );
    }

    /// Return attribute with given name (or null if such an attribute does not
    /// exist).
    public Attribute getAttribute( string name ) {
	foreach( Attribute a in attrs )
	    if( a.name == name )
		return a;
	return null;
    }

    /// Return all methods that have the given attribute.
    public IList attrMethods( string attr ) {
	IList result = new ArrayList( 5 );
	foreach( Method m in methods )
	    if( m.getAttribute( attr ) != null )
		result.Add( m );
	return result;
    }
	     
    //
    // constructor handling
    //

    /*** 
     * Call the constructor that matches given arguments (which must be proper
     * values). Throws exception if arguments do not match any constructor. 
     * @see startInputIteration
     */
    public void construct( IList args ) { 
#if DEBUG
	Console.WriteLine( "constructing " + name + "(" + args + ")" );
#endif
	Class oldCurrClass = currClass;
	currClass = this;

	// execute initialization statements
	if( initStatements != null )
	    foreach( Statement s in initStatements )
		s.execute();

	// no constructor?
	if( ctors.Count == 0 ) {
	    if( args.Count == 0 ) {
		currClass = oldCurrClass;
#if DEBUG
		Console.WriteLine( "constructed " + name + "(" + args + ")" );
		dumpState();
#endif
		return;
	    }
	} else 
	    foreach( Method c in ctors ) {
		if( c.compatibleArgs( args ) ) {
		    c.execute( args );
		    currClass = oldCurrClass;
#if DEBUG
		    Console.WriteLine( "constructed " + name + "(" + args + ")" );
		    dumpState();
#endif
		    return;
		}
	    }

	throw new Exception( "Class.construct(): given arguments do not match any constructor" );
    }


    /// Return whether any class ctor is callable with arguments args (which
    /// must be either proper values or Expr derivatives).
    public bool constructible( IList args ) {
	// allow no constructor
	if( ctors.Count == 0 && args.Count == 0 )
	    return true;
	foreach( Method c in ctors )
	    if( c.compatibleArgs( args ) )
		return true;
	return false;
    }

    public override string ToString() {
	string result = "class " + name + ": " + parent + " " +
	    "[" + attrs + "] {\n";
	foreach( Method c in ctors )
	    result += "    ctor " + c + "\n";

	foreach( Method m in methods )
	    result += "    " + m + "\n";

	foreach( DictionaryEntry i in fields )
	    result += "    " + i.Value + "\n";

	return result + "}";
    }

#if DEBUG
    /// Write the value of the class fields to the standard output (for
    /// debugging purposes).
    public void dumpState() {
	foreach( DictionaryEntry i in fields )
	    Console.WriteLine( "   " + i.Value );
    }
#endif

    /**
     * Set this class as the current class. This must be done when evaluating
     * the possible values of a stub property, to allow reading the value of a
     * stub field for dynamic bounds.
     * @see Expr.doID
     */
    public void becomeCurrentClass() { currClass = this; }

    /// Return the currently active class.
    public static Class currentClass() { return currClass; }

    //
    // state retrieval / restauration
    //

    /**
     * Create a state process node from the current field values. Stub
     * references are stored as a string containing the class name, to avoid
     * serializing classes.
     */
    public StateNode getState() {
	IList state = new ArrayList();
	Object o;
	foreach( DictionaryEntry i in fields ) {
	    // ignore Comp.active field
	    if( ( (Variable) i.Value ).name == "active" )
		continue;
	    o = ( (Variable) i.Value ).value;
	    if( o is Class )
		state.Add( ( (Class) o ).getName() );
	    else if( o is List ) 
		state.Add( ( (List) o ).DeepClone() );
	    else
		state.Add( o );
	}
	return new StateNode( state );
    }

    /** 
     * Set the fields to the values given in the state node. Does not replace
     * Stub references since only their class names are stored in the
     * StateNode.
     */
    public void setState( StateNode sn ) {
	IDictionaryEnumerator i = fields.GetEnumerator();
	Variable field;
	foreach( Object o in sn.getState() ) {
	    i.MoveNext();
	    field = (Variable) i.Value;
	    // ignore Comp.active field
	    if( field.name == "active" ) {
		i.MoveNext();
		field = (Variable) i.Value;
	    }
	    if( !( field.value is Class ) ) {
		if( o is List )
		    field.value = ( (List) o ).DeepClone();
		else
		    field.value = o;
	    }
	}
    }

    /// Return a list of variable names which describes the format of generated
    /// state nodes.
    public IList stateFormat() {
	List names = new List();
	foreach( DictionaryEntry i in fields )
	    if( ( (string) i.Key ) != "active" )
		names.Add( i.Key );
	return names;
    }

    /// Calculate state abstraction equivalence relation from attributes.
    public StateAbstraction getStateAbstraction() {
	int index = -1;
	string abstracts;
	Variable v;
	EqRel eq;
	// Count-1 because we don't count "active"
	StateAbstraction sa = new StateAbstraction( fields.Count-1 );

	foreach( DictionaryEntry i in fields ) {
	    v = (Variable) i.Value;
	    if( v.name == "active" )
		continue;
	    index++;

	    // ignore stub fields
	    if( v.type == this.GetType() )
		continue;

	    // determine abstracts
	    Attribute abstrAttr = v.getAttribute( "Abstracts" );
	    if( abstrAttr == null )
		abstracts = v.name;
	    else {
		if( abstrAttr.positional.Count == 1 && abstrAttr.positional[0] is string )
		    abstracts = (string) abstrAttr.positional[0];
		else
		    throw new Exception( "invalid Abstracts attribute (must contain only a string)" );
	    }

	    // create equivalence releation from Eq attributes
	    eq = new EqRel();
	    List attrs = v.getAttributes();
	    if( attrs != null )
		foreach( Attribute a in attrs )
		    if( a.name == "Eq" )
			parseEqAttr( a, eq );

	    sa.addField( index, abstracts, eq );
	}

	return sa;
    }

    /// Calculate transition abstraction equivalence relation from attributes.
    public TransitionAbstraction getTransitionAbstraction() {
	TransitionAbstraction ta = new TransitionAbstraction();
	string abstracts;
	Variable v;

	foreach( DictionaryEntry i in fields ) {
	    v = (Variable) i.Value;
	    if( v.name == "active" )
		continue;

	    // ignore non-stub fields
	    if( v.type != this.GetType() )
		continue;

	    // determine abstracts
	    Attribute abstrAttr = v.getAttribute( "Abstracts" );
	    if( abstrAttr == null )
		abstracts = v.name;
	    else {
		if( abstrAttr.positional.Count == 1 && abstrAttr.positional[0] is string )
		    abstracts = (string) abstrAttr.positional[0];
		else
		    throw new Exception( "invalid Abstracts attribute (must contain only a string)" );
	    }

	    EqRel eq = null;
	    Hashtable propRels = new Hashtable(); // string -> EqRel

	    // determine equivalence relations for each of the properties
	    List attrs = v.getAttributes();
	    if( attrs == null )
		continue;
	    foreach( Attribute a in attrs ) {
		// Property attribute creates new EqRel
		if( a.name == "Property" ) {
		    eq = new EqRel();
		    if( a.positional.Count == 1 && a.positional[0] is string )
			propRels[a.positional[0]] = eq;
		    else
			throw new Exception( "invalid Property attribute (must contain only a string)" );
		}

		if( a.name == "Eq" ) {
		    if( eq == null )
			throw new Exception( "Eq attributes of stub fields must be preceeded by a Property attribute" );
		    parseEqAttr( a, eq );
		}
	    }

	    ta.addStub( v.name, abstracts, propRels );
	}

	return ta;
    }

    /// Parse given Eq attribute and update given EqRel accordingly.
    private void parseEqAttr( Attribute a, EqRel eq ) {
	if( a.name != "Eq" ) 
	    return;

	if( a.positional.Count != 2 || a.named.Count != 0 )
	    throw new Exception( "Invalid Eq attribute " + a + 
		    ": must contain exactly two positional arguments" );

	if( a.positional[1] is string ) {
	    if( fields.ContainsKey( a.positional[1] ) )
		eq.addSingle( a.positional[0], ( (Variable) fields[a.positional[1]] ).value );
	    else
		eq.addRange( a.positional[0], parseRange( (string) a.positional[1] ) );
	} else
	    eq.addSingle( a.positional[0], a.positional[1] );
    }

    /// Parse given range string.
    private Range parseRange( string r ) {
	// construct regular expression for all valid ranges
	Regex regex = new Regex("^([\\(\\[])\\s*(-?\\d*\\.\\d+|-?\\d+|-oo|[a-zA-Z_]\\w*)\\s*,\\s*(-?\\d*\\.\\d+|-?\\d+|\\+oo|[a-zA-Z_]\\w*)\\s*([\\)\\]])$");
	Match m = regex.Match( r );
	if( m.Success )
	    return new Range( m.Groups[1].Value == "(", 
		    parseBound( m.Groups[2].Value ), 
		    m.Groups[4].Value == ")", 
		    parseBound( m.Groups[3].Value ) );
	else
	    throw new Exception( "invalid range format: " + r );
    }

    private Object parseBound( string bound ) {
	try {
	    return bool.Parse( bound );
	} catch( FormatException ) {
	    try {
		return int.Parse( bound );
	    } catch( FormatException ) {
		try {
		    return double.Parse( bound );
		} catch( FormatException ) {
		    if( bound == "-oo" || bound == "+oo" || bound == "oo" )
			return null;
		    Variable v = (Variable) fields[bound];
		    if( v != null )
			return v.value;
		    else
			throw new Exception( "invalid range: field " + bound +
				" does not exist" );
		}
	    }
	}
    }

    /// Add the CCM specific members for class "Comp".
    private void addCCMCompMembers() {
	addField( new Variable( "active", false.GetType(), null, false, null, false ) );
	// add EventOccured()
	Method m = new Method( "EventOccured", null, new List(0), null, false );
	m.setBody( new Block( null ) );
	addMethod( m );
    }

    // Add the CCM specific members for class "Stub".
    private void addCCMStubMembers() {
	Method m = new Method( "acquireCooperation", null, new List(0), null, false );
	m.setBody( new Block( null ) );
	addMethod( m );

	m = new Method( "releaseCooperation", null, new List(0), null, false );
	m.setBody( new Block( null ) );
	addMethod( m );
    }

    string name, parent;
    IList ctors; // of Method (only void return type)
    IList methods; // of Method
    Hashtable fields; // string -> Variable
    List attrs; // of Attribute
    IList initStatements; // of Statement: commands to initialize fields
    static Class currClass = null; // expressions need this to call methods

    /// True if this is a Stub class, false in other classes.
    bool isStub;

}

} // closing namespace
