/**
 * @file decls.cs 
 * Classes to represent attributes, qualified identifiers, variables and
 * fields.
 *
 * (C) 2004 Martin Pitt <martin@piware.de>
 *
 * This software is distributed under the terms and conditions of the GNU
 * General Public License (see file COPYING)
 */

using System;
using System.Collections;

namespace CCM {

/// Represents a qualified identifier.
public class QID {
    public string id;
    public QID next; // null if last component

    /// Creates the initial part of a QID with given identifier.
    public QID( string id ) {
	this.id = id;
	this.next = null;
    }

    /// Append an identifier.
    public void add( string q ) {
	if( next == null )
	    next = new QID( q );
	else
	    next.add( q );
    }

    /// Returns a C# conformant string representation with dot separators.
    public override string ToString() {
	if( next == null )
	    return id;
	else
	    return id + "." + next;
    }
}

/**
 * Represents a variable or method reference. To be able to resolve static
 * method calls, you must call setClasses() before use.
 */
public class Ref {
    /** 
     * Initializes a reference with the given id. If thisRef == true, it
     * becomes a reference to this.id. 
     */
    public Ref( QID id, bool thisRef ) {
	this.id = id;
	this.thisRef = thisRef;
    }

    /// Return whether the reference is qualified.
    public bool isQualified() { return id.next != null; }

    /// Returns the qualified identifer of this reference.
    public QID getId() { return id; }

    /// Returns whether this is a "this" reference.
    public bool isThis() { return thisRef; }

    /**
     * Get/set the value of the referred variable. This also ensures that
     * ValueInNodes and ValueOutNodes are created on stub accesses.
     */
    public Object Value {
	get {
	    if( isQualified() )
		return StubIterator.stubInput( this );
	    else {
		Variable v = resolveVar();
		return v.value;
	    }
	}

	set {
	    if( isQualified() )
		StubIterator.stubOutput( this, value );
	    else {
		Variable v = resolveVar();
		if( (value != null ) && (
			    v.array && !( value is List ) ||
			    !v.array && value.GetType() != v.type )
		  )
		    throw new Exception( "Ref.Value setting: types do not match: "
			    + v.type + " != " + value.GetType() );

		v.value = value;
	    }
	}
    }

    /** 
     * Call the method this reference points to and return its result. This
     * also ensures that a MethodCallNode is created on stub method calls.
     */
    public Object call( List args ) {
	// ignore Common.log and such things
	if( id.id == "Common" )
	    return null;
	if( isQualified() )
	    return StubIterator.call( this, args );
	else
	    return Class.currentClass().getMethod( id.id, args ).execute( args );
    }

    /// Returns a C# representation of this reference.
    public override string ToString() {
	if( thisRef )
	    return "this." + id;
	else
	    return id.ToString();
    }

    /// This must be called with a dictionary of all available classes before
    /// code interpretation to allow resolution of static method calls.
    public static void setClasses( Hashtable classes ) {
	Ref.classes = classes;
    }

    /** 
     * Returns the variable this reference points to. This does not yet
     * support static class fields! 
     */
    private Variable resolveVar() {
	Variable v = null;

	if( !thisRef ) {
	    v = Block.getCurrentBlock().getVar( id.id ); 
	    if( v == null )
		v = Method.getCallParam( id.id );
	    if( v != null )
		return v;
	}

	v = Class.currentClass().getField( id.id );
	if( v == null )
	    throw new Exception( "Ref.resolve(): unknown variable: " + id.id + 
		    " in class " + Class.currentClass().getName() );

	if( id.next != null ) {
	    v = v.stubClass.getField( id.next.id );
	    if( v == null )
		throw new Exception( "Ref.resolve(): unknown stub field: " + id.next.id + 
			" in class " + v.stubClass.getName() );
	}

	return v;
    }

    /** Return the stub class of the referred variable. Also works for static
      method call references. */
    public Class resolveClass() {
	// static members
	Class c = (Class) classes[id.id];
	if( c != null )
	    return c;

	Variable v = Class.currentClass().getField( id.id );
	if( v == null )
	    throw new Exception( "Ref.resolveClass(): unknown variable: " + id.id + 
		    " in class " + Class.currentClass().getName() );
	if( v.value != null )
	    return (Class) v.value;
	else
	    return v.stubClass;
    }

    private QID id;

    /// True if this is a this.xxx reference.
    private bool thisRef; 

    /// string -> Class dictionary of all available classes for static method
    /// call resolution.
    private static Hashtable classes; 
}

/// Represents an attribute with positional and named arguments.
public class Attribute {
    /// Creates an attribute with the given positional and named arguments.
    public Attribute( string name, List positional, IDictionary named ) {
	this.name = name;
	this.positional = positional;
	this.named = named;
    }

    /// Returns a C# representation of this attribute.
    public override string ToString() {
	string result = name + "(" + positional;

	if( named != null && named.Count > 0 )
	    foreach( DictionaryEntry i in named )
		result += ", " + i.Key + " = " + i.Value;
	return result + ")";
    }

    public string name;
    public List positional;
    public IDictionary named;
}

/**
 * Represents a typed variable declaration with attributes. This is mainly used
 * for storing method parameters. This class implements IEnumerable to be able
 * to iterate over all possible values.
 */
public class VariableDecl : IEnumerable {
    /** 
     * Creates an uninitialized (null) variable or class field.
     * @param name Variable name
     * @param type Variable type (e. g. 0.GetType() )
     * @param stubClass If type is Class.GetType(), then the variable is a stub
     *        reference and stubClass the concrete stub class
     * @param array If true, then this variable is an array of type
     * @param attrs List of Attributes
     * @param _public Whether this is a public field
     */
    public VariableDecl( string name, Type type, Class stubClass, bool array, List attrs, bool _public ) {
	this.name = name;
	this.type = type;
	this.array = array;
	this.attrs = attrs;
	this.stubClass = stubClass;
	this._public = _public;
    }

    public override string ToString() {
	string result;
	if( attrs != null && attrs.Count > 0 )
	    result = "[" + attrs +  "] ";
	else
	    result = "";
	string t;
	if( stubClass != null )
	    t = stubClass.getName();
	else
	    t = type.ToString();
	if( array )
	    return result + t + "[] " + name;
	else
	    return result + t + " " + name;
    }

    /// Return an enumerator over all possible values.
    public IEnumerator GetEnumerator() {
	if( type == false.GetType() )
	    return new BoolIterator();
	if( type == 0.GetType() )
	    return new IntIterator( getAttribute( "Range" ) );
	if( type == (0.0d).GetType() )
	    return new DoubleIterator( getAttribute( "Range" ) );

	throw new Exception( "VariableDecl.GetEnumerator(): cannot iterate over type " + type );
    }

    /// Return the first attribute with given name, or null if no such
    /// attribute exists.
    public Attribute getAttribute( string name ) {
	if( attrs == null )
	    return null;
	foreach( Attribute a in attrs )
	    if( a.name == name )
		return a;
	return null;
    }

    /// Return a List of all attributes or null if no attributes exist.
    public List getAttributes() {
	return attrs;
    }

    public string name;
    public Type type;
    public Class stubClass; // if Stub type
    public bool array;
    public bool _public;
    protected List attrs;
}

/// Represents a variable or a class field.
public class Variable : VariableDecl {
    /** 
     * Creates an uninitialized (null) variable or class field.
     * @param name Variable name
     * @param type Variable type (e. g. 0.GetType() )
     * @param stubClass If type is Class.GetType(), then the variable is a stub
     *        reference and stubClass the concrete stub class
     * @param array If true, then this variable is an array of type
     * @param attrs List of Attributes
     * @param _public Whether this is a public field
     */
    public Variable( string name, Type type, Class stubClass, bool array, List attrs, bool _public ) 
	: base( name, type, stubClass, array, attrs, _public ) {
	if( type == 0.GetType() )
	    this.value = 0;
	else if( type == (0.0d).GetType() )
	    this.value = 0.0d;
	else if( type == "".GetType() )
	    this.value = "";
	else if( type == false.GetType() )
	    this.value = false;
	else
	    this.value = null;
    }

    /**
     * Clones the given variable.
     */
    public Variable( Variable v ) 
	: base( v.name, v.type, v.stubClass, v.array, v.attrs, v._public ) {
	this.value = v.value;
    }

    /** 
     * Creates an initialized variable or class field.
     * @param name Variable name
     * @param type Variable type (e. g. 0.GetType() )
     * @param stubClass If type is Class.GetType(), then the variable is a stub
     *        reference and stubClass the concrete stub class
     * @param array If true, then this variable is an array of type
     * @param attrs List of Attributes
     * @param _public Whether this is a public field
     * @param value Initial value
     */
    public Variable( string name, Type type, Class stubClass, bool array, List attrs, bool _public, Object value ) 
	: base( name, type, stubClass, array, attrs, _public ) {
	if( value is List )
	    this.value = ( (List) value ).DeepClone();
	else
	    this.value = value;
    }

    public override string ToString() {
	if( array )
	    return base.ToString() + " = " + 
		(value == null ? "null" : ("["+value+"]") );
	else
	    return base.ToString() + " = " + 
		(value == null ? "null" : value.ToString() );
    }

    /** Current variable value. Arrays are stored as List of values. */
    public Object value = null;
}

/**
 * List that provides a nice ToString(). Otherwise it behaves like ArrayList.
 */
[Serializable]
public class List : ArrayList {
    /// Creates an empty list with a default initial capacity.
    public List() : base() {}

    /// Creates an empty list with given initial capacity.
    public List( int capacity ) : base( capacity ) {}

    /// Return a deep copy of the List (concerning contained Lists)
    public List DeepClone() {
	List result = new List( Count );
	foreach( Object o in this )
	    if( o is List )
		result.Add( ( (List) o ).DeepClone() );
	    else
		result.Add( o );
	return result;
    }

    /// Checks whether this list and list l contain the same elements.
    public override bool Equals( Object lobj ) {
	List l = lobj as List;
	if( l == null || l.Count != Count )
	    return false;
	IEnumerator le = l.GetEnumerator();
	foreach( Object o in this ) {
	    le.MoveNext();
	    if( !o.Equals( le.Current ) )
		return false;
	}
	return true;
    }

    /// Formats the elements as a comma separated list.
    public override string ToString() {
	string result = "";
	bool first = true;
	foreach( Object o in this ) {
	    if( first )
		first = false;
	    else
		result += ", ";
	    if( o == null )
		result += "null";
	    else if( o is string )
		result += "\"" + o +"\"";
	    else if( o is Class )
		result += ( (Class) o ).getName();
	    else
		result += o;
	}
	return result;
    }
}

} // closing namespace
