/**
 * @file eqrel.cs 
 * Classes to deal with abstraction: value range, equivalence relation, strate
 * and transition abstraction relation.
 *
 * (C) 2004 Martin Pitt <martin@piware.de>
 *
 * This software is distributed under the terms and conditions of the GNU
 * General Public License (see file COPYING)
 */

using System;
using System.IO;
using System.Collections;
using System.Runtime.Serialization;

namespace CCM {

/// Value range for equivalence relation domains.
[Serializable]
public class Range {
    /**
     * Create an interval range.
     * @param leftopen True if left interval border is open
     * @param left int or double value of the left (minimum) value; null means
     *        "minus infinity" (only for leftopen=true)
     * @param rightopen True if left interval border is open
     * @param right Value of the right (maximum) value (same type as left); null means
     *        "plus infinity" (only for rightopen=true)
     */
    public Range( bool leftopen, Object left, bool rightopen, Object right ) {
	if( !leftopen && left == null )
	    throw new Exception( "Invalid interval: closed minus infinite left border" );
	if( !rightopen && right == null )
	    throw new Exception( "Invalid interval: closed infinite right border" );

	if( !( left == null || left is int || left is double ) )
	    throw new Exception( "Left value must be null, int or double");
	if( right != null && left != null && right.GetType() != left.GetType() )
	    throw new Exception( "Right value must be null or same type as left");

	this.leftopen = leftopen;
	this.left = left;
	this.rightopen = rightopen;
	this.right = right;

	if( left != null && right != null && compare( left, right ) == 1 )
	    throw new Exception( "Invalid interval: minimal bound greater than maximal bound" );
    }

    /// Returns whether given value is within the range.
    public bool contains( Object value ) {
	int cmp;

	if( left != null ) {
	    cmp = compare( left, value );
	    if( cmp >= ( leftopen ? 0 : 1 ) )
		return false;
	}

	if( right != null ) {
	    cmp = compare( value, right );
	    if( cmp >= ( rightopen ? 0 : 1 ) )
		return false;
	}

	return true;
    }

    public override string ToString() {
	return ( leftopen ? "(" : "[" ) +
	    ( left == null ? "-oo" : left ) + 
	    ", " + 
	    ( right == null ? "+oo" : right ) + 
	    ( rightopen ? ")" : "]" );
    }

    /**
     * Compare the two objects (int or double).
     * @return -1 if o1<o2, 0 if o1==o2, 1 if o1>o2
     */
    private static int compare( Object o1, Object o2 ) {
	if( o1.Equals( o2 ) )
	    return 0;

	if( o1 is int && o2 is int )
	    return ( ( (int) o1 ) < ( (int) o2 ) ) ? -1 : 1;

	double v1 = (o1 is int) ? ( (int) o1 ) : ( (double) o1 );
	double v2 = (o2 is int) ? ( (int) o2 ) : ( (double) o2 );
	return v1 < v2 ? -1 : 1;
    }

    private Object left, right;
    private bool leftopen, rightopen;
}

/**
 * Equivalence relation for a single value.
 */
[Serializable]
public class EqRel {
    /// Create new equivalence relation.
    public EqRel() {
	eq = new Hashtable();
    }

    /**
     * Add single value entry.
     * @param abstr Abstract value
     * @param impl Equivalent implementation value
     */
    public void addSingle( Object abstr, Object impl ) {
	if( abstr == null )
	    throw new Exception( "abstract value must not be null" );
	if( impl == null )
	    throw new Exception( "implementation value must not be null" );
	if( eq.ContainsKey( abstr ) )
	    throw new Exception( "abstract value " + abstr + " already exists in relation" );
	eq[abstr] = impl;
    }

    /**
     * Add equivalence class entry.
     * @param value Abstract value
     * @param r equivalence class that val abstracts
     */
    public void addRange( Object value, Range r ) {
	if( eq[value] == null )
	    eq[value] = new List();
	( (List) eq[value] ).Add( r );
    }

    /// Return whether two values are related. 
    public bool equal( Object o1, Object o2 ) {
	if( o1 == null )
	    return o2 == null;
	Object entry = eq[o1];

	/// directly compare objects if no equivalence class exists
	if( entry == null )
	    return o1.Equals( o2 );

	if( entry is List ) {
	    if( o2 == null )
		return false;
	    foreach( Range r in (List) entry )
		if( r.contains( o2 ) )
		    return true;
	    return false;
	} else
	    return entry.Equals( o2 );
    }

    public override string ToString() {
	string result = "EqRel { ";
	foreach( DictionaryEntry i in eq )
	    result += i.Key + " -> " + i.Value+ ", ";
	return result + "}";
    }

    /**
     * Stores the relation. Maps Object to a List of Ranges to
     * accomodate non-contiguous equivalence classes.
     */
    private Hashtable eq; 
}

/// State abstraction (equivalence relation for StateNodes). 
[Serializable]
public class StateAbstraction {
    /// Create a new state abstraction relation for numFields abstraction class fields.
    public StateAbstraction( int numFields ) {
	abstractsIndex = null;
	rel = new EqRel[numFields];
	abstracts = new string[numFields];
    }

    /**
     * Add new field to the end of the array.
     * @param index Index of this field in the StateNode state list given to
     *        equal()
     * @param abstracts Name of the respective implementation field
     * @param eq EqRel of this field
     */
    public void addField( int index, string abstracts, EqRel eq ) {
	if( index >=  rel.Length )
	    throw new Exception( "index of new field outside of the range given in the ctor" );
	this.abstracts[index] = abstracts;
	this.rel[index] = eq;
    }

    /**
     * Finishes adding fields and resolves implementation names to their index
     * in implementation StateNodes. Resets abstracts to null.
     * @param implNames Layout of implementation StateNodes (contains field
     *        names), as delivered by Class.stateName()
     * @see Class.stateName()
     */
    public void resolve( IList implNames ) {
	abstractsIndex = new int[abstracts.Length];

	int idx;
	for( int i = 0; i < abstracts.Length; ++i ) {
	    // fields that are not added are not compared (e. g. stub fields)
	    if( abstracts[i] == null ) {
		abstractsIndex[i] = -1;
		continue;
	    }

	    idx = implNames.IndexOf( abstracts[i] );
	    if( idx == -1 && abstracts[i] != "active" )
		throw new Exception( "given list of implementation field names does not contain " +
			abstracts[i] );
	    abstractsIndex[i] = idx;
	}
	
	abstracts = null;
    }

    /**
     * Return whether the two given StateNodes are equivalent. The state
     * abstraction must be resolve()d for this.
     * @see resolve
     */
    public bool equal( StateNode abstr, StateNode impl ) {
	if( abstracts != null )
	    throw new Exception( "equal() called without prior call to resolve()" );
	int implIndex;

	IList astates = abstr.getState();
	IList istates = impl.getState();

	// handle stop state "0"
	if( astates == null || istates == null )
	    return ( astates == null && istates == null );

	for( int i = 0; i < astates.Count; ++i ) {
	    implIndex = abstractsIndex[i];
	    if( implIndex > 0 && 
		    !rel[i].equal( astates[i], istates[implIndex] ) )
		return false;
	}

	return true;
    }

    public override string ToString() {
	string result;
	if( abstracts == null ) {
	    result = "Resolved StateAbstraction {\n";
	    for( int i = 0; i < abstractsIndex.Length; ++i )
		if( abstractsIndex[i] != -1 )
		    result += "  " + i + "->" + abstractsIndex[i] + ": " + rel[i] + "\n";
	    result += "}\n";
	} else {
	    result = "Unresolved StateAbstraction {\n";
	    for( int i = 0; i < abstracts.Length; ++i )
		if( abstracts[i] != null )
		    result += "  " + i + "->" + abstracts[i] + ": " + rel[i] + "\n";
	    result += "}\n";
	}
	return result;
    }

    /// Maps state array indices of the abstraction to the field name of the
    /// implementation.
    private string[] abstracts;

    /// Maps state array indices of the abstraction to an EqRel.
    private EqRel[] rel;

    /**
     * Maps state array indices of the abstraction to the respective state
     * array index of the implementation. Created by resolve() from abstracts.
     */
    private int[] abstractsIndex;
}

/// Transition abstraction (equivalence relation for TransitionNodes).
[Serializable]
public class TransitionAbstraction {
    /// Create a new transition abstraction relation.
    public TransitionAbstraction() {
	rel = new Hashtable();
	abstracts = new Hashtable();
    }

    /**
     * Add new stub field.
     * @param name Name of the abstract stub field
     * @param abstracts Name of the respective implementation stub field
     * @param eq Hashtable (string -> EqRel) that maps stub properties and
     *        methods to their equivalence relations
     */
    public void addStub( string name, string abstracts, Hashtable eq ) {
	if( this.abstracts.ContainsKey( name ) )
	    throw new Exception( "Relations for stub field "+ name+" already assigned" );
	this.abstracts[name] = abstracts;
	rel[name] = eq;
    }

    /// Return whether the two given TransitionNodes are equivalent.
    public bool equal( TransitionNode abstr, TransitionNode impl ) {
	if( abstr.GetType() != impl.GetType() )
	    return false;

	// check callargs equality for MethodCallNodes
	if( abstr is MethodCallNode ) {
	    IList aargs = ( (MethodCallNode) abstr ).getCallArgs();
	    IList iargs = ( (MethodCallNode) impl ).getCallArgs();
	    if( aargs.Count != iargs.Count )
		return false;

	    IEnumerator ienum = iargs.GetEnumerator();
	    foreach( Object arg in aargs ) {
		ienum.MoveNext();
		if( arg == null ) 
		    return ienum.Current == null;
		if( !arg.Equals( ienum.Current ) )
		    return false;
	    }
	}

	// determine abstract and implementation field
	string abstrField = abstr.getField();
	string implField = abstrField;
	if( abstrField != null ) {
	    implField = (string) abstracts[abstrField];
	    if( implField == null )
		implField = abstrField;
	}

	// compare class, field and property name
	if( implField != impl.getField() ||
	    abstr.getProp() != impl.getProp() )
	    return false;

	Object av = abstr.getValue();
	Object iv = impl.getValue();
	if( av == null )
	    return iv == null;

	// compare value with equivalence relation or directly (when no EqRel
	// is available)
	EqRel eq = null;
	if( abstrField != null ) {
	    Hashtable stubrels = (Hashtable) rel[abstrField];
	    if( stubrels == null )
		return av.Equals( iv );
	    eq = (EqRel) stubrels[abstr.getProp()];
	}

	if( eq == null ) {
	    return av.Equals( iv );
	} else
	    return eq.equal( av, iv );
    }

    public override string ToString() {
	string result = "TransitionAbstraction {\n";
	foreach( DictionaryEntry r in rel ) {
	    result += "  " + r.Key + " -> " + abstracts[r.Key] + " {\n";
	    foreach( DictionaryEntry s in (Hashtable) r.Value )
		result += "    " + s.Key + " -> " + s.Value + "\n";
	    result += "  }\n";
	}
	return result + "}\n";
    }

    /// Maps abstraction stub field names to implementation stub field names
    /// (string -> string).
    private Hashtable abstracts;

    /// Maps abstraction stub field names to a hashtable of stub properties to
    /// equivalence releation (string -> (string -> EqRel)).
    private Hashtable rel;
}
}
