/**
 * @file expr.cs
 * Classes to represent and evaluate a C# expression tree.
 *
 * (C) 2004 Martin Pitt <martin@piware.de>
 *
 * This software is distributed under the terms and conditions of the GNU
 * General Public License (see file COPYING)
 */

using System;
using System.Collections;

namespace CCM {


/**
 * Base class of all expressions. If an expression contains stub property
 * accesses or stub method calls, there are several possible results which must
 * be iterated through in the caller. 
 */
public abstract class Expr {

    /** 
     * Evaluate expression and return its value. Stub access expressions add
     * transitions to the currently evaluated process.
     */
    abstract public Object eval();

    /**
     * Return the expression type. This is generally the type of the expression
     * value, not of the Expr itself (e. g. "int", not "IntExpr").
     */
    abstract public Type type();

    /**
     * Call method and return its value (for a method call expression). Common
     * for all expression types.
     */
    protected Object doCall() {
	List args = new List( callargs.Count );
	foreach( Expr e in callargs )
	    args.Add( e.eval() );
	return id.call( args );
    }

    /**
     * Return the array element for an element expression. Common for all
     * expression types.
     */
    protected Object doElemAccess() {
	return ( (ArrayExpr) left ).getElement( (int) index.eval() );
    }

    /**
     * Execute assignmet for an array element assignment expression. Common for
     * all expression types.
     */
    protected Object doElemAssignment() {
	Object value = right.eval();
	( (ArrayExpr) left ).setElement( (int) index.eval(), right.eval() );
	return value;
    }

    /// Return whether this is an assignable expression (array element or ID).
    public bool assignable() { return (index != null) || ( id != null && callargs == null ) ;}

    /// Return the referred identifier for reference expressions.
    public Ref getRef() { return id; }

    /// Return whether this is an element access expression.
    public bool isElementAccess() { return index != null; }

    /// Return an expression that assigns value to the element if called on an
    /// array element access expression. 
    abstract public Expr makeElementAssign( Expr value );

    /// Reference for a variable/stub property reference expressions.
    protected Ref id;

    /// Left subexpression for binary expressions, subexpression for unary
    /// expressions.
    protected Expr left; 
    
    /// Right subexpression for binary expressions, unused for unary
    /// expressions.
    protected Expr right; 

    /// Element index for array element access expressions.
    protected IntExpr index;

    /// Argument list for method call expressions (of Expr).
    protected List callargs; 
}

/**
 * Expression without type. The only valid instance is "null".
 */
public class NullExpr : Expr {
    public override Object eval() {
	return null;
    }

    public override Type type() {
	return null;
    }

    public override Expr makeElementAssign( Expr value ) {
	throw new Exception( "NullExpr.makeElementAssign: not allowed" );
    }
}

/// Boolean type expression
public class BoolExpr : Expr {
    /// Possible operations.
    public enum Op { LITERAL, ID, CALL, NOT, EQ, NEQ, AND, OR, LT, LE, GT, GE, ASSIGN, ELEMENT, SETELEMENT };

    /// Create a literal expression with given value.
    public BoolExpr( bool lit ) {
	op = Op.LITERAL;
	value = lit;
    }

    /// Create a variable / stub property reference expression.
    public BoolExpr( Ref id ) {
	op = Op.ID;
	this.id = id;
    }

    /**
     * Create assignment expression.
     * @param id Reference to assign to
     * @param e Expression to assign to id
     */
    public BoolExpr( Ref id, Expr e ) {
	op = Op.ASSIGN;
	this.id = id;
	left = e;
    }

    /// Create method call expression to method(args).
    public BoolExpr( Ref method, List args ) {
	op = Op.CALL;
	id = method;
	callargs = args;
    }

    /// Create array element access expression to array[index].
    public BoolExpr( ArrayExpr array, IntExpr index ) {
	op = Op.ELEMENT;
	left = array;
	this.index = index;
    }

    /// Create array element assignment expression: array[index] = value.
    public BoolExpr( ArrayExpr array, IntExpr index, Expr value ) {
	op = Op.SETELEMENT;
	left = array;
	right = value;
	this.index = index;
    }

    /// Create unary expression.
    public BoolExpr( Op op, Expr e ) {
	this.op = op;
	left = e;
    }

    /// Create binary expression.
    public BoolExpr( Op op, Expr e1, Expr e2 ) {
	this.op = op;
	left = e1;
	right = e2;
    }

    public override Type type() { return false.GetType(); }

    public override Object eval()
    {
	switch( op ) {
	    case Op.LITERAL:	return value;
	    case Op.ID:		return id.Value;
	    case Op.ASSIGN:	return id.Value = left.eval();
	    case Op.CALL:	return doCall();
	    case Op.ELEMENT:	return doElemAccess();
	    case Op.SETELEMENT:	return doElemAssignment();

	    case Op.NOT:	return ! (bool) left.eval();

            case Op.EQ:
                if( left is NullExpr )
                    return right.eval() == null;
                if( right is NullExpr )
                    return left.eval() == null;
                return left.eval().Equals( right.eval() );

            case Op.NEQ:
                if( left is NullExpr )
                    return right.eval() != null;
                if( right is NullExpr )
                    return left.eval() != null;
                return !( left.eval().Equals( right.eval() ) );

	    case Op.AND:	return (bool) left.eval() && (bool) right.eval();
	    case Op.OR:		return (bool) left.eval() || (bool) right.eval();

	    case Op.LT:
		if( left is IntExpr )
		    return (int) left.eval() < (int) right.eval();
		else
		    return (double) left.eval() < (double) right.eval();
	    case Op.LE:
		if( left is IntExpr )
		    return (int) left.eval() <= (int) right.eval();
		else
		    return (double) left.eval() <= (double) right.eval();
	    case Op.GT:
		if( left is IntExpr )
		    return (int) left.eval() > (int) right.eval();
		else
		    return (double) left.eval() > (double) right.eval();
	    case Op.GE:
		if( left is IntExpr )
		    return (int) left.eval() >= (int) right.eval();
		else
		    return (double) left.eval() >= (double) right.eval();
	}
	return null;
    }

    public override Expr makeElementAssign( Expr value ) {
	if( index == null )
	    throw new Exception( "BoolExpr.makeElementAssign: this expression is not an element access" );
	return new BoolExpr( (ArrayExpr) left, index, value );
    }

    public override string ToString()
    {
	switch( op ) {
	    case Op.LITERAL:	return value.ToString();
	    case Op.ID:		return id.ToString();
	    case Op.NOT:	return "!" + left;

	    case Op.EQ:		return "("+left+"=="+right+")";
	    case Op.NEQ:	return "("+left+"!="+right+")";
	    case Op.AND:	return "("+left+"&&"+right+")";
	    case Op.OR:		return "("+left+"||"+right+")";

	    case Op.LT:		return "("+left+"<"+right+")";
	    case Op.LE:		return "("+left+"<="+right+")";
	    case Op.GT:		return "("+left+">"+right+")";
	    case Op.GE:		return "("+left+">="+right+")";

	    case Op.ASSIGN:	return "("+id+"="+left+")";
	    case Op.CALL:	return id + "(" + callargs + ")";
	    case Op.ELEMENT:	return left + "[" + index + "]";
	    case Op.SETELEMENT:	return "("+left + "[" + index + "]"+"="+right+")";
	}
	return null;
    }

    Op op;
    bool value;
}

/// Integer typ expression.
public class IntExpr : Expr {
    /// Possible operations.
    public enum Op { 
	LITERAL, ID, CALL, ELEMENT, SETELEMENT,
	TILDE, NEG, 
	ADD, SUB, MUL, DIV, MOD, SHL, SHR, AND, XOR, OR, ASSIGN };

    /// Create a literal expression with given value.
    public IntExpr( int lit ) {
	op = Op.LITERAL;
	value = lit;
    }

    /// Create a variable / stub property reference expression.
    public IntExpr( Ref id ) {
	op = Op.ID;
	this.id = id;
    }

    /**
     * Create assignment expression.
     * @param id Reference to assign to
     * @param e Expression to assign to id
     */
    public IntExpr( Ref id, Expr e ) {
	op = Op.ASSIGN;
	this.id = id;
	left = e;
    }

    /// Create method call expression to method(args).
    public IntExpr( Ref method, List args ) {
	op = Op.CALL;
	id = method;
	callargs = args;
    }

    /// Create array element access expression to array[index].
    public IntExpr( ArrayExpr array, IntExpr index ) {
	op = Op.ELEMENT;
	left = array;
	this.index = index;
    }

    /// Create array element assignment expression: array[index] = value.
    public IntExpr( ArrayExpr array, IntExpr index, Expr value ) {
	op = Op.SETELEMENT;
	left = array;
	right = value;
	this.index = index;
    }

    /// Create unary expression.
    public IntExpr( Op op, Expr e ) {
	this.op = op;
	left = e;
    }

    /// Create binary expression.
    public IntExpr( Op op, Expr e1, Expr e2 ) {
	this.op = op;
	left = e1;
	right = e2;
    }

    public override Type type() { return 0.GetType(); }

    public override Object eval() 
    {
	switch( op ) {
	    case Op.LITERAL:	return value;
	    case Op.ID:		return id.Value;
	    case Op.ASSIGN:	return id.Value = left.eval();
	    case Op.CALL:	return doCall();
	    case Op.ELEMENT:	return doElemAccess();
	    case Op.SETELEMENT:	return doElemAssignment();

	    case Op.TILDE:	return ~ (int) left.eval();
	    case Op.NEG:	return - (int) left.eval();

	    case Op.ADD:	return (int) left.eval() + (int) right.eval();
	    case Op.SUB:	return (int) left.eval() - (int) right.eval();
	    case Op.MUL:	return (int) left.eval() * (int) right.eval();
	    case Op.DIV:	return (int) left.eval() / (int) right.eval();
	    case Op.MOD:	return (int) left.eval() % (int) right.eval();
	    case Op.SHL:	return (int) left.eval() << (int) right.eval();
	    case Op.SHR:	return (int) left.eval() >> (int) right.eval();
	    case Op.AND:	return (int) left.eval() & (int) right.eval();
	    case Op.XOR:	return (int) left.eval() ^ (int) right.eval();
	    case Op.OR:		return (int) left.eval() | (int) right.eval();
	}
	return null;
    }

    public override Expr makeElementAssign( Expr value ) {
	if( index == null )
	    throw new Exception( "IntExpr.makeElementAssign: this expression is not an element access" );
	return new IntExpr( (ArrayExpr) left, index, value );
    }

    public override string ToString()
    {
	switch( op ) {
	    case Op.LITERAL:	return value.ToString();
	    case Op.ID:		return id.ToString();
	    case Op.TILDE:	return "~" + left;
	    case Op.NEG:	return "-" + left;

	    case Op.ADD:	return "("+left+"+"+right+")";
	    case Op.SUB:	return "("+left+"-"+right+")";
	    case Op.MUL:	return "("+left+"*"+right+")";
	    case Op.DIV:	return "("+left+"/"+right+")";
	    case Op.MOD:	return "("+left+"%"+right+")";
	    case Op.SHL:	return "("+left+"<<"+right+")";
	    case Op.SHR:	return "("+left+">>"+right+")";
	    case Op.AND:	return "("+left+"&"+right+")";
	    case Op.XOR:	return "("+left+"^"+right+")";
	    case Op.OR:		return "("+left+"|"+right+")";

	    case Op.ASSIGN:	return "("+id+"="+left+")";
	    case Op.CALL:	return id + "(" + callargs + ")";
	    case Op.ELEMENT:	return left + "[" + index + "]";
	    case Op.SETELEMENT:	return "("+left + "[" + index + "]"+"="+right+")";
	}
	return null;
    }

    Op op;
    int value;
}


/// Float type expression.
public class FloatExpr : Expr {
    /// Possible operations.
    public enum Op { LITERAL, ID, CALL, NEG, ADD, SUB, MUL, DIV, ASSIGN, ELEMENT, SETELEMENT };

    /// Create a literal expression with given value.
    public FloatExpr( double lit ) {
	op = Op.LITERAL;
	value = lit;
    }

    /// Create a variable / stub property reference expression.
    public FloatExpr( Ref id ) {
	op = Op.ID;
	this.id = id;
    }

    /**
     * Create assignment expression.
     * @param id Reference to assign to
     * @param e Expression to assign to id
     */
    public FloatExpr( Ref id, Expr e ) {
	op = Op.ASSIGN;
	this.id = id;
	left = e;
    }

    /// Create method call expression to method(args).
    public FloatExpr( Ref method, List args ) {
	op = Op.CALL;
	id = method;
	callargs = args;
    }

    /// Create array element access expression to array[index].
    public FloatExpr( ArrayExpr array, IntExpr index ) {
	op = Op.ELEMENT;
	left = array;
	this.index = index;
    }

    /// Create array element assignment expression: array[index] = value.
    public FloatExpr( ArrayExpr array, IntExpr index, Expr value ) {
	op = Op.SETELEMENT;
	left = array;
	right = value;
	this.index = index;
    }

    /// Create unary expression.
    public FloatExpr( Op op, Expr e ) {
	this.op = op;
	left = e;
    }

    /// Create binary expression.
    public FloatExpr( Op op, Expr e1, Expr e2 ) {
	this.op = op;
	left = e1;
	right = e2;
    }

    public override Type type() { return 0.0d.GetType(); }

    public override Object eval() 
    {
	switch( op ) {
	    case Op.LITERAL:	return value;
	    case Op.ID:		return id.Value;
	    case Op.ASSIGN:	return id.Value = left.eval();
	    case Op.CALL:	return doCall();
	    case Op.ELEMENT:	return doElemAccess();
	    case Op.SETELEMENT:	return doElemAssignment();

	    case Op.NEG:	return - (double) left.eval();
	    case Op.ADD:	return (double) left.eval() + (double) right.eval();
	    case Op.SUB:	return (double) left.eval() - (double) right.eval();
	    case Op.MUL:	return (double) left.eval() * (double) right.eval();
	    case Op.DIV:	return (double) left.eval() / (double) right.eval();
	}
	return null;
    }

    public override Expr makeElementAssign( Expr value ) {
	if( index == null )
	    throw new Exception( "FloatExpr.makeElementAssign: this expression is not an element access" );
	return new FloatExpr( (ArrayExpr) left, index, value );
    }

    public override string ToString()
    {
	switch( op ) {
	    case Op.LITERAL:	return value.ToString();
	    case Op.ID:		return id.ToString();
	    case Op.NEG:	return "-" + left;

	    case Op.ADD:	return "("+left+"+"+right+")";
	    case Op.SUB:	return "("+left+"-"+right+")";
	    case Op.MUL:	return "("+left+"*"+right+")";
	    case Op.DIV:	return "("+left+"/"+right+")";

	    case Op.ASSIGN:	return "("+id+"="+left+")";
	    case Op.CALL:	return id + "(" + callargs + ")";
	    case Op.ELEMENT:	return left + "[" + index + "]";
	    case Op.SETELEMENT:	return "("+left + "[" + index + "]"+"="+right+")";
	}
	return null;
    }

    Op op;
    double value;
}

/// String type expression.
public class StringExpr : Expr {
    /// Possible operations.
    public enum Op { LITERAL, ID, CALL, CONCAT, ASSIGN, ELEMENT, SETELEMENT }

    /// Create a literal expression with given value.
    public StringExpr( string lit ) {
	this.op = Op.LITERAL;
	value = lit;
    }

    /// Create a variable / stub property reference expression.
    public StringExpr( Ref id ) {
	this.op = Op.ID;
	this.id = id;
    }

    /**
     * Create assignment expression.
     * @param id Reference to assign to
     * @param e Expression to assign to id
     */
    public StringExpr( Ref id, Expr e ) {
	op = Op.ASSIGN;
	this.id = id;
	left = e;
    }

    /// Create method call expression to method(args).
    public StringExpr( Ref method, List args ) {
	op = Op.CALL;
	id = method;
	callargs = args;
    }

    /// Create array element access expression to array[index].
    public StringExpr( ArrayExpr array, IntExpr index ) {
	op = Op.ELEMENT;
	left = array;
	this.index = index;
    }

    /// Create array element assignment expression: array[index] = value.
    public StringExpr( ArrayExpr array, IntExpr index, Expr value ) {
	op = Op.SETELEMENT;
	left = array;
	right = value;
	this.index = index;
    }

    /// Create binary expression.
    public StringExpr( Op op, Expr e1, Expr e2 ) {
	this.op = op;
	left = e1;
	right = e2;
    }

    public override Type type() { return "".GetType(); }

    public override Object eval() 
    {
	switch( op ) {
	    case Op.LITERAL:	return value;
	    case Op.ID:		return id.Value;
	    case Op.ASSIGN:	return id.Value = left.eval();
	    case Op.CALL:	return doCall();
	    case Op.ELEMENT:	return doElemAccess();
	    case Op.SETELEMENT:	return doElemAssignment();
	    case Op.CONCAT:	return (string) left.eval() + (string) right.eval();
	}
	return null; // not reached
    }

    public override Expr makeElementAssign( Expr value ) {
	if( index == null )
	    throw new Exception( "StringExpr.makeElementAssign: this expression is not an element access" );
	return new StringExpr( (ArrayExpr) left, index, value );
    }

    public override string ToString()
    {
	switch( op ) {
	    case Op.LITERAL:	return "\"" + value + "\"";
	    case Op.ID:		return id.ToString();
	    case Op.CONCAT:	return "("+left+"+"+right+")";
	    case Op.ASSIGN:	return "("+id+"="+left+")";
	    case Op.CALL:	return id + "(" + callargs + ")";
	    case Op.ELEMENT:	return left + "[" + index + "]";
	    case Op.SETELEMENT:	return "("+left + "[" + index + "]"+"="+right+")";
	}
	return null; // not reached
    }

    Op op;
    string value;
}

/// Array type expression.
public class ArrayExpr : Expr {
    /// Possible operations.
    public enum Op { LITERAL, NEW, ID, ASSIGN, ELEMENT, SETELEMENT };

    /// Create a variable / stub property reference expression.
    public ArrayExpr( Ref id, Type elemType ) {
	op = Op.ID;
	this.id = id;
	this.elemType = elemType;
    }

    /// Create new expression: new elemType[size].
    public ArrayExpr( Type elemType, IntExpr size ) {
	op = Op.NEW;
	this.elemType = elemType;
	left = size;
    }

    /**
     * Create assignment expression.
     * @param id Reference to assign to
     * @param e Expression to assign to id
     * @param elemType Type of the contained elements
     */
    public ArrayExpr( Ref id, Expr e, Type elemType ) {
	op = Op.ASSIGN;
	this.id = id;
	left = e;
	this.elemType = elemType;
    }

    /// Create a literal expression with given value.
    public ArrayExpr( List lit, Type elemType ) {
	op = Op.LITERAL;
	value = lit;
	this.elemType = elemType;
    }

    /// Create array element access expression to array[index].
    public ArrayExpr( ArrayExpr array, IntExpr index ) {
	op = Op.ELEMENT;
	left = array;
	this.index = index;
    }

    /// Create array element assignment expression: array[index] = value.
    public ArrayExpr( ArrayExpr array, IntExpr index, Expr value ) {
	op = Op.SETELEMENT;
	left = array;
	right = value;
	this.index = index;
    }

    public override Type type() { return ( new List() ).GetType(); }

    /// Return the type of contained elements.
    public Type elementType() { return elemType; }

    /**
     * Return the value of thisarray[index].
     * @see Expr.doElemAccess
     */
    public Object getElement( int index ) {
	List val = (List) this.eval();
	if( val == null )
	    throw new Exception( "ArrayExpr.getElement(): " + ToString() + " = null" ); 
	if( index < 0 || index >= val.Count )
	    throw new Exception( "ArrayExpr.getElement(): index " + index + 
		    " is out of array bounds [0, " + (val.Count-1) + "]" );
	return val[index];
    }

    /**
     * Set the value of thisarray[index] = value.
     * @see Expr.doElemAssignment
     */
    public void setElement( int index, Object value ) {
	List val = (List) this.eval();
	if( val == null )
	    throw new Exception( "ArrayExpr.setElement(): " + ToString() + " = null" ); 
	if( index < 0 || index >= val.Count )
	    throw new Exception( "ArrayExpr.setElement(): index " + index + 
		    " is out of array bounds [0, " + (val.Count-1) + "]" );
	val[index] = value;
    }

    public override Object eval() 
    {
	switch( op ) {
	    case Op.LITERAL:	return value;
	    case Op.ID:		return id.Value;
	    case Op.ASSIGN:	return id.Value = left.eval();
	    case Op.ELEMENT:	return doElemAccess();
	    case Op.SETELEMENT:	return doElemAssignment();

	    case Op.NEW:
		Object init = null;
		if( elemType == 0.GetType() )
		    init = 0;
		else if( elemType == (0.0d).GetType() )
		    init = 0.0d;
		else if( elemType == false.GetType() )
		    init = false;
		else if( elemType == "".GetType() )
		    init = "";
		int size = (int) left.eval();
		value = new List( size );
		for( int i = 0; i < size; ++i )
		    value.Add( init );
		return value;
	}
	
	return null; // not reached
    }

    public override Expr makeElementAssign( Expr value ) {
	if( index == null )
	    throw new Exception( "ArrayExpr.makeElementAssign: this expression is not an element access" );
	return new ArrayExpr( (ArrayExpr) left, index, value );
    }

    public override string ToString()
    {
	switch( op ) {
	    case Op.LITERAL:	return "{" + value + "}";
	    case Op.ID:		return id.ToString();
	    case Op.ASSIGN:	return "("+id+"="+left+")";
	    case Op.NEW:	return "new " + elemType + "[" + left.eval() + "]";
	    case Op.ELEMENT:	return left + "[" + index + "]";
	    case Op.SETELEMENT:	return "("+left + "[" + index + "]"+"="+right+")";
	}
	return null; // not reached
    }

    Op op;
    List value;
    Type elemType;
}


/// Stub expression.
public class StubExpr : Expr {
    /// Possible operations.
    public enum Op { NEW, ID, ASSIGN, CALL, ELEMENT, SETELEMENT };

    /// Create a variable / stub property reference expression.
    public StubExpr( Ref id ) {
	op = Op.ID;
	this.id = id;
    }

    /// Create new expression: new stub( args ).
    public StubExpr( Class stub, List args ) {
	op = Op.NEW;
	stubclass = stub;
	callargs = args;
    }

    /**
     * Create assignment expression.
     * @param id Reference to assign to
     * @param e Expression to assign to id
     */
    public StubExpr( Ref id, Expr e ) {
	op = Op.ASSIGN;
	this.id = id;
	left = e;
    }

    /// Create array element access expression to array[index].
    public StubExpr( ArrayExpr array, IntExpr index ) {
	op = Op.ELEMENT;
	left = array;
	this.index = index;
    }

    /// Create array element assignment expression: array[index] = value.
    public StubExpr( ArrayExpr array, IntExpr index, Expr value ) {
	op = Op.SETELEMENT;
	left = array;
	right = value;
	this.index = index;
    }

    /// Create method call expression to method(args).
    public StubExpr( Ref id, List args ) {
	op = Op.CALL;
	this.id = id;
	callargs = args;
    }

    public override Type type() { 
	return (new Class( "", null, "" ) ).GetType();
    }

    public override Object eval() {
	Class c;
	switch( op ) {
	    case Op.ID:	    return id.Value;
	    case Op.ASSIGN: return id.Value = left.eval();
	    case Op.CALL:   return doCall();
	    case Op.ELEMENT:return doElemAccess();
	    case Op.SETELEMENT:	return doElemAssignment();

	    case Op.NEW:    
		c = new Class( stubclass );
		List args = new List( callargs.Count );
		foreach( Expr e in callargs )
		    args.Add( e.eval() );
		c.construct( args );
		return c;
	}

	return null; // not reached
    }

    public override Expr makeElementAssign( Expr value ) {
	if( index == null )
	    throw new Exception( "StubExpr.makeElementAssign: this expression is not an element access" );
	return new StubExpr( (ArrayExpr) left, index, value );
    }

    public override string ToString() {
	switch( op ) {
	    case Op.ID:		return id.ToString();
	    case Op.ASSIGN:	return "("+id + "=" + left + ")";
	    case Op.CALL:	return id + "(" + callargs + ")"; 
	    case Op.NEW:	return stubclass.getName() + "(" + callargs + ")"; 
	    case Op.ELEMENT:	return left + "[" + index + "]";
	    case Op.SETELEMENT:	return "("+left + "[" + index + "]"+"="+right+")";
	}

	return null; // not reached
    }

    Op op;
    Class stubclass; // for NEW
}

/// "this" expression.
public class ThisExpr : Expr {
    /// Create "this" expression referring to given class.
    public ThisExpr( Class c ) {
	cls = c;
    }

    public override Object eval() {
	return cls;
    }

    public override Type type() { return this.GetType(); }

    public override Expr makeElementAssign( Expr value ) {
	throw new Exception( "ThisExpr.makeElementAssign: not allowed" );
    }

    public override string ToString() {
	return "this";
    }

    Class cls;
}


/// Void expression for void method calls.
public class VoidExpr : Expr {
    /// Create void method call expression: method( args ).
    public VoidExpr( Ref method, List args ) {
	id = method;
	callargs = args;
    }

    public override Type type() { return null; }

    public override Object eval()
    {
	doCall();
	return null;
    }

    public override Expr makeElementAssign( Expr value ) {
	throw new Exception( "VoidExpr.makeElementAssign: not allowed" );
    }

    public override string ToString() {
        return id + "(" + callargs + ")";
    }
}

} // closing namespace
